<!doctype html>
<html>
  <head>
    <title>{{ page_title }} &mdash; Cool Magic!</title>
    <link rel="stylesheet" href="{{ h.url_for('shared_data', file='style.css') }}" type="text/css">
  </head>
  <body>
    <h1>Cool Magic</h1>
    <h2>{{ page_title }}</h2>
    {% block page_body %}{% endblock %}
  </body>
</html>
