<!doctype html>
<html>
  <head>
    <title>{{ title }} | Example Application</title>
  </head>
  <body>
    <h1>Example Application</h1>
    <p>
      Request Language: <strong>{{ req.language }}</strong>
    </p>
    {% block body %}{% endblock %}
    <blockquote>
      <p>This page in other languages:
      <ul>
      {% for lng in ['en', 'de', 'fr'] %}
        <li><a href="{{ req.url_for('this', lang_code=lng) }}">{{ lng }}</a></li>
      {% endfor %}
      </ul>
    </blockquote>
  </body>
</html>
