<!doctype html>
<html>
  <head>
    <title>Teeworlds Server Browser</title>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
    <link rel="stylesheet" type="text/css" href="{{ url_for('shared', file='style.css') }}">
  </head>
  <body>
    <h1><a href="https://www.teeworlds.com/">Teeworlds</a> Server Browser</h1>
    <div class="contents">
      {% block body %}{% endblock %}
    </div>
    <div class="footer">
      <a href="https://www.teeworlds.com/">Teeworlds</a> Server Browser by
      <a href="http://lucumr.pocoo.org/">Armin Ronacher</a>, powered by
      <a href="http://werkzeug.pocoo.org/">Werkzeug</a>, licensed under
      the modified BSD license.
    </div>
  </body>
</html>
